@@ -8,17 +8,21 @@ require Exporter;
 
 @ISA = qw(Exporter);
 
-$VERSION = '1.84';
+$VERSION = '1.86';
 
 %EXPORT_TAGS =
 (
 	ALL =>		[ qw( classical inflect
 			      PL PL_N PL_V PL_ADJ NO NUM A AN
 			      PL_eq PL_N_eq PL_V_eq PL_ADJ_eq
+			      PART_PRES
+			      ORD
+			      NUMWORDS
 		     	      def_noun def_verb def_adj def_a def_an )],
 
 	INFLECTIONS =>	[ qw( classical inflect
-			      PL PL_N PL_V PL_ADJ PL_eq NO NUM A AN )],
+			      PL PL_N PL_V PL_ADJ PL_eq
+			      NO NUM A AN PART_PRES )],
 
 	PLURALS =>	[ qw( classical inflect
 			      PL PL_N PL_V PL_ADJ NO NUM
@@ -29,6 +33,8 @@ $VERSION = '1.84';
 
 	ARTICLES =>	[ qw( classical inflect NUM A AN )],
 
+	NUMERICAL =>	[ qw( ORD NUMWORDS )],
+
 	USER_DEFINED =>	[ qw( def_noun def_verb def_adj def_a def_an )],
 );
 
@@ -45,7 +51,7 @@ sub classical
 
 my $persistent_count;
 
-sub NUM(;$$)	# (;$count,$show)
+sub NUM		# (;$count,$show)
 {
 	if (defined $_[0])
 	{
@@ -62,7 +68,7 @@ sub NUM(;$$)	# (;$count,$show)
 
 # 0. PERFORM GENERAL INFLECTIONS IN A STRING
 
-sub inflect($)
+sub inflect
 {
 	my $save_persistent_count = $persistent_count;
 	my @sections = split /(NUM\([^)]*\))/, $_[0];
@@ -76,8 +82,11 @@ sub inflect($)
 		       s/\bPL   \( ([^),]*) (, ([^)]*) )? \)  / PL($1,$3)   /xeg
 		    || s/\bPL_N \( ([^),]*) (, ([^)]*) )? \)  / PL_N($1,$3) /xeg
 		    || s/\bPL_V \( ([^),]*) (, ([^)]*) )? \)  / PL_V($1,$3) /xeg
+		    || s/\bPL_ADJ \( ([^),]*) (, ([^)]*) )? \)  / PL_ADJ($1,$3) /xeg
 		    || s/\bAN?  \( ([^),]*) (, ([^)]*) )? \)  / A($1,$3)    /xeg
 		    || s/\bNO   \( ([^),]*) (, ([^)]*) )? \)  / NO($1,$3)   /xeg
+		    || s/\bORD  \( ([^)]*) \)                 / ORD($1)   /xeg
+		    || s/\bNUMWORDS  \( ([^)]*) \)                 / NUMWORDS($1)   /xeg
 		}
 
 		$inflection .= $_;
@@ -130,15 +139,17 @@ my $PL_sb_irregular = join '|', keys %PL_sb_irregular;
 
 # CLASSICAL "..a" -> "..ata"
 
-my $PL_sb_C_a_ata = join "|", map { chop; $_; }
+my @PL_sb_C_a_ata = 
 (
 	"anathema", "bema", "carcinoma", "charisma", "diploma",
 	"dogma", "drama", "edema", "enema", "enigma", "lemma",
 	"lymphoma", "magma", "melisma", "miasma", "oedema",
 	"sarcoma", "schema", "soma", "stigma", "stoma", "trauma",
-	"gumma",
+	"gumma", "pragma",
 );
 
+my $PL_sb_C_a_ata = join "|", map { substr($_,0,-1) } @PL_sb_C_a_ata;
+
 # UNCONDITIONAL "..a" -> "..ae"
 
 my $PL_sb_U_a_ae = join "|", 
@@ -152,19 +163,19 @@ my $PL_sb_C_a_ae = join "|",
 (
 	"amoeba", "antenna", "formula", "hyperbola",
 	"medusa", "nebula", "parabola", "abscissa",
-	"hydra", "nova", "lacuna", "aurora",
+	"hydra", "nova", "lacuna", "aurora", ".*umbra",
 );
 
 # CLASSICAL "..en" -> "..ina"
 
-my $PL_sb_C_en_ina = join "|", map { chop; chop; $_; }
+my $PL_sb_C_en_ina = join "|", map { substr($_,0,-2) }
 (
 	"stamen",	"foramen",	"lumen",
 );
 
 # UNCONDITIONAL "..um" -> "..a"
 
-my $PL_sb_U_um_a = join "|", map { chop; chop; $_; }
+my $PL_sb_U_um_a = join "|", map { substr($_,0,-2) }
 (
 	"bacterium",	"agendum",	"desideratum",	"erratum",
 	"stratum",	"datum",	"ovum",		"extremum",
@@ -173,7 +184,7 @@ my $PL_sb_U_um_a = join "|", map { chop; chop; $_; }
 
 # CLASSICAL "..um" -> "..a"
 
-my $PL_sb_C_um_a = join "|", map { chop; chop; $_; }
+my $PL_sb_C_um_a = join "|", map { substr($_,0,-2) }
 (
 	"maximum",	"minimum",	"momentum",	"optimum",
 	"quantum",	"cranium",	"curriculum",	"dictum",
@@ -187,7 +198,7 @@ my $PL_sb_C_um_a = join "|", map { chop; chop; $_; }
 
 # UNCONDITIONAL "..us" -> "i"
 
-my $PL_sb_U_us_i = join "|", map { chop; chop; $_; }
+my $PL_sb_U_us_i = join "|", map { substr($_,0,-2) }
 (
 	"alumnus",	"alveolus",	"bacillus",	"bronchus",
 	"locus",	"nucleus",	"stimulus",	"meniscus",
@@ -195,7 +206,7 @@ my $PL_sb_U_us_i = join "|", map { chop; chop; $_; }
 
 # CLASSICAL "..us" -> "..i"
 
-my $PL_sb_C_us_i = join "|", map { chop; chop; $_; }
+my $PL_sb_C_us_i = join "|", map { substr($_,0,-2) }
 (
 	"focus",	"radius",	"genius",
 	"incubus",	"succubus",	"nimbus",
@@ -214,7 +225,7 @@ my $PL_sb_C_us_us = join "|",
 
 # UNCONDITIONAL "..on" -> "a"
 
-my $PL_sb_U_on_a = join "|", map { chop; chop; $_; }
+my $PL_sb_U_on_a = join "|", map { substr($_,0,-2) }
 (
 	"criterion",	"perihelion",	"aphelion",
 	"phenomenon",	"prolegomenon",	"noumenon",
@@ -223,7 +234,7 @@ my $PL_sb_U_on_a = join "|", map { chop; chop; $_; }
 
 # CLASSICAL "..on" -> "..a"
 
-my $PL_sb_C_on_a = join "|", map { chop; chop; $_; }
+my $PL_sb_C_on_a = join "|", map { substr($_,0,-2) }
 (
 	"oxymoron",
 );
@@ -235,7 +246,7 @@ my @PL_sb_C_o_i =
 	"solo",		"soprano",	"basso",	"alto",
 	"contralto",	"tempo",
 );
-my $PL_sb_C_o_i = join "|", map { my $w=$_; chop $w; $w; } @PL_sb_C_o_i;
+my $PL_sb_C_o_i = join "|", map { substr($_,0,-1) } @PL_sb_C_o_i;
 
 # ALWAYS "..o" -> "..os"
 
@@ -258,14 +269,14 @@ my $PL_sb_U_o_os = join "|",
 
 # UNCONDITIONAL "..ex" -> "..ices"
 
-my $PL_sb_U_ex_ices = join "|", map { chop; chop; $_; }
+my $PL_sb_U_ex_ices = join "|", map { substr($_,0,-2) }
 (
 	"codex",	"murex",	"silex",
 );
 
 # CLASSICAL "..ex" -> "..ices"
 
-my $PL_sb_C_ex_ices = join "|", map { chop; chop; $_; }
+my $PL_sb_C_ex_ices = join "|", map { substr($_,0,-2) }
 (
 	"vortex",	"vertex",	"cortex",	"latex",
 	"pontifex",	"apex",		"index",	"simplex",
@@ -288,8 +299,12 @@ my $PL_sb_C_im = join "|",
 # UNCONDITIONAL "..man" -> "..mans"
 
 my $PL_sb_U_man_mans = join "|", 
-(
-	"german", "human",
+qw(
+	human
+	Alabaman Bahaman Burman German
+	Hiroshiman Liman Nakayaman Oklahoman
+	Panaman Selman Sonaman Tacoman Yakiman
+	Yokohaman Yuman
 );
 
 my @PL_sb_uninflected_s =
@@ -374,7 +389,7 @@ my $PL_prep = join '|', qw (
         since till to under until unto upon with
 );
 
-my $PL_sb_prep_compound = '(.*?)((-|\s+)(in|to|of|at|de)(-|\s+)(.*))';
+my $PL_sb_prep_compound = '(.*?)((-|\s+)('.$PL_prep.'|d[eu])((-|\s+)(.*))?)';
 
 
 my %PL_pron_nom =
@@ -519,39 +534,39 @@ my @PL_v_user_defined  = ();
 my @PL_adj_user_defined  = ();
 my @A_a_user_defined   = ();
 
-sub def_noun($$)
+sub def_noun
 {
-unshift @PL_sb_user_defined, checkpatsubs(@_);
-return 1;
+	unshift @PL_sb_user_defined, checkpatsubs(@_);
+	return 1;
 }
 
-sub def_verb($$$$$$)
+sub def_verb
 {
-unshift @PL_v_user_defined, checkpatsubs(@_[4,5]);
-unshift @PL_v_user_defined, checkpatsubs(@_[2,3]);
-unshift @PL_v_user_defined, checkpatsubs(@_[0,1]);
-return 1;
+	unshift @PL_v_user_defined, checkpatsubs(@_[4,5]);
+	unshift @PL_v_user_defined, checkpatsubs(@_[2,3]);
+	unshift @PL_v_user_defined, checkpatsubs(@_[0,1]);
+	return 1;
 }
 
-sub def_adj($$)
+sub def_adj
 {
-unshift @PL_adj_user_defined, checkpatsubs(@_);
-return 1;
+	unshift @PL_adj_user_defined, checkpatsubs(@_);
+	return 1;
 }
 
-sub def_a($)	
+sub def_a
 {
 unshift @A_a_user_defined, checkpat(@_,'a');
 return 1;
 }
 
-sub def_an($)
+sub def_an
 {
 unshift @A_a_user_defined, checkpat(@_,'an');
 return 1;
 }
 
-sub ud_match($@)
+sub ud_match
 {
 my $word = shift;
 for (my $i=0; $i < @_; $i+=2)
@@ -559,7 +574,7 @@ for (my $i=0; $i < @_; $i+=2)
 	if ($word =~ /^(?:$_[$i])$/i)
 	{
 		last unless defined $_[$i+1];
-		return eval '"'.$_[$i+1].' $word"';
+		return eval '"'.$_[$i+1].'"';
 	}
 }
 return undef;
@@ -580,7 +595,7 @@ do $rcfile or die "\nBad .inflectrc file ($rcfile):\n\t$@\n"
 if $rcfile && -r $rcfile && -s $rcfile;
 };
 
-sub postprocess($$)	# FIX PEDANTRY AND CAPITALIZATION :-)
+sub postprocess		# FIX PEDANTRY AND CAPITALIZATION :-)
 {
 my ($orig, $inflected) = @_;
 $inflected =~ s/([^|]+)\|(.+)/ $classical?$2:$1 /e;
@@ -590,7 +605,7 @@ return $orig =~ /^I$/	? $inflected
  :			  $inflected;
 }
 
-sub PL($;$)
+sub PL
 #   PL($word,$number)
 {
 my ($str, $count) = @_;
@@ -602,7 +617,7 @@ my $plural = postprocess $word,  _PL_special_adjective($word,$count)
 return $pre.$plural.$post;
 }
 
-sub PL_N($;$)	
+sub PL_N
 #   PL_N($word,$number)
 {
 my ($str, $count) = @_;
@@ -612,7 +627,7 @@ my $plural = postprocess $word, _PL_noun($word,$count);
 return $pre.$plural.$post;
 }
 
-sub PL_V($;$)	
+sub PL_V
 #   PL_V($word,$number)
 {
 my ($str, $count) = @_;
@@ -623,7 +638,7 @@ my $plural = postprocess $word, _PL_special_verb($word,$count)
 return $pre.$plural.$post;
 }
 
-sub PL_ADJ($;$)	
+sub PL_ADJ
 #   PL_ADJ($word,$number)
 {
 my ($str, $count) = @_;
@@ -634,12 +649,12 @@ my $plural = postprocess $word, _PL_special_adjective($word,$count)
 return $pre.$plural.$post;
 }
 
-sub PL_eq($$)	  { _PL_eq(@_, \&PL); }
-sub PL_N_eq($$)	  { _PL_eq(@_, \&PL_N); }
-sub PL_V_eq($$)	  { _PL_eq(@_, \&PL_V); }
-sub PL_ADJ_eq($$) { _PL_eq(@_, \&PL_ADJ); }
+sub PL_eq	  { _PL_eq(@_, \&PL); }
+sub PL_N_eq	  { _PL_eq(@_, \&PL_N); }
+sub PL_V_eq	  { _PL_eq(@_, \&PL_V); }
+sub PL_ADJ_eq	  { _PL_eq(@_, \&PL_ADJ); }
 
-sub _PL_eq($$$)
+sub _PL_eq
 {
 my ( $word1, $word2, $PL ) = @_;
 my $classval = $classical;
@@ -668,12 +683,12 @@ if ($PL == \&PL || $PL == \&PL_ADJ)
 return $result;
 }
 
-sub _PL_reg_plurals($$$$)
+sub _PL_reg_plurals
 {
 $_[0] =~ /($_[1])($_[2]\|\1$_[3]|$_[3]\|\1$_[2])/
 }
 
-sub _PL_check_plurals_N($$)
+sub _PL_check_plurals_N
 {
 my $pair = "$_[0]|$_[1]";
 foreach ( values %PL_sb_irregular_s )	{ return 1 if $_ eq $pair; }
@@ -699,7 +714,7 @@ return 1 if _PL_reg_plurals($pair, $PL_sb_C_a_ata,   "as","ata")
 return 0;
 }
 
-sub _PL_check_plurals_ADJ($$$)
+sub _PL_check_plurals_ADJ
 {
 my ( $word1a, $word2a ) = @_;
 my ( $word1b, $word2b ) = @_;
@@ -728,7 +743,7 @@ if ($word1b)
 return "";
 }
 
-sub _PL_noun($;$)
+sub _PL_noun
 {
 my ( $word, $count ) = @_;
 my $value;				# UTILITY VARIABLE
@@ -845,7 +860,8 @@ $word =~ /o$/i			and return "${word}es";
 
 $word =~ /^(?:$PL_sb_general)$/i
 				and $value = $2
-				and return _PL_noun($1,2,"$1s")
+				# and return _PL_noun($1,2,"$1s")
+				and return _PL_noun($1,2)
 					   . $value;
 
 $word =~ /^(?:$PL_sb_prep_compound)$/i
@@ -859,7 +875,7 @@ return "${word}s";
 }
 
 
-sub _PL_special_verb($;$)
+sub _PL_special_verb
 {
 my ( $word, $count ) = @_;
 $count = $persistent_count
@@ -893,7 +909,7 @@ $word =~ /\s/				and return undef;
 
 # HANDLE STANDARD 3RD PERSON (CHOP THE ...(e)s OFF SINGLE WORDS)
 
-$word =~ /^(.*)([cs]h|[sx]|zz)es$/i	and return "$1$2";
+$word =~ /^(.*)([cs]h|[x]|zz|ss)es$/i	and return "$1$2";
 
 $word =~ /^(..+)ies$/i			and return "$1y";
 $word =~ /^(.+)oes$/i			and return "$1o";
@@ -905,7 +921,7 @@ $word =~ /^(.*[^s])s$/i			and return $1;
 return undef;
 }  
 
-sub _PL_general_verb($;$)
+sub _PL_general_verb
 {
 my ( $word, $count ) = @_;
 $count = $persistent_count
@@ -932,7 +948,7 @@ return $word;
 
 }
 
-sub _PL_special_adjective($;$)
+sub _PL_special_adjective
 {
 my ( $word, $count ) = @_;
 $count = $persistent_count
@@ -996,7 +1012,7 @@ my $A_explicit_an = join '|',
 "hour(?!i)", "heir", "honest", "hono",
 );
 
-sub A($;$) 
+sub A
 {
 my ($str, $count) = @_;
 my ($pre, $word, $post) = ( $str =~ m/\A(\s*)(.+?)(\s*)\Z/ );
@@ -1005,9 +1021,9 @@ my $result = _indef_article($word,$count);
 return $pre.$result.$post;
 }
 
-sub AN($;$) { goto &A }
+sub AN { goto &A }
 
-sub _indef_article($;$)
+sub _indef_article
 {
 my ( $word, $count ) = @_;
 
@@ -1057,7 +1073,7 @@ $word =~ /^($A_y_cons)/io	and return "an $word";
 
 # 2. TRANSLATE ZERO-QUANTIFIED $word TO "no PL($word)"
 
-sub NO($;$)
+sub NO
 {
 my ($str, $count) = @_;
 my ($pre, $word, $post) = ($str =~ m/\A(\s*)(.+?)(\s*)\Z/);
@@ -1068,7 +1084,210 @@ $count = 0 unless $count;
 
 return "$pre$count " . PL($word,$count) . $post
 	unless $count =~ /^$PL_count_zero$/;
-return "${pre}no ". PL($word,0) . $post;
+return "${pre}no ". PL($word,0) . $post ;
+}
+
+
+# PARTICIPLES
+
+sub PART_PRES
+{
+        local $_ = PL_V(shift,2);
+           s/ie$/y/
+        or s/ue$/u/
+        or s/([auy])e$/$1/
+        or s/i$//
+        or s/([^e])e$/$1/
+        or m/er$/
+        or s/([^aeiou][aeiouy]([bdgmnprst]))$/$1$2/;
+        return "${_}ing";
+}
+
+
+
+# NUMERICAL INFLECTIONS
+
+my %nth =
+(
+	0 => 'th',
+	1 => 'st',
+	2 => 'nd',
+	3 => 'rd',
+	4 => 'th',
+	5 => 'th',
+	6 => 'th',
+	7 => 'th',
+	8 => 'th',
+	9 => 'th',
+	11 => 'th',
+	12 => 'th',
+	13 => 'th',
+);
+
+
+sub ORD
+{
+	$_[0] . ($nth{$_[0]%100} || $nth{$_[0]%10});
+}
+
+
+my %default_args = 
+(
+	'group'   => 0,
+	'comma'   => ',',
+	'and'     => 'and',
+	'zero'    => 'zero',
+	'decimal' => 'point',
+);
+
+my @unit = ('',qw(one two three four five six seven eight nine));
+my @teen = qw(ten eleven twelve thirteen fourteen
+	      fifteen sixteen seventeen eighteen nineteen);
+my @ten  = ('','',qw(twenty thirty forty fifty sixty seventy eighty ninety));
+my @mill = map { (my $val=$_) =~ s/_/illion/; " $val" }
+	   ('',qw(thousand m_ b_ tr_ quadr_ quint_ sext_ sept_ oct_ non_ dec_));
+
+sub mill { my $ind = $_[0]||0;
+	   die "Number out of range\n" if $ind > $#mill;
+	   return $ind<@mill ? $mill[$ind] : ' ???illion'; }
+
+sub unit { return $unit[$_[0]]. mill($_[1]); }
+
+sub ten
+{
+	return $ten[$_[0]] . ($_[0]&&$_[1]?'-':'') . $unit[$_[1]] . mill($_[2])
+		if $_[0] ne '1';
+	return $teen[$_[1]]. $mill[$_[2]||0];
+}
+
+sub hund
+{
+	return unit($_[0]) . " hundred" . ($_[1] || $_[2] ? " $_[4] " : '')
+	     . ten($_[1],$_[2]) . mill($_[3]) . ', ' if $_[0];
+	return ten($_[1],$_[2]) . mill($_[3]) . ', ' if $_[1] || $_[2];
+	return '';
+}
+
+sub enword
+{
+	my ($num,$group,$zero,$comma,$and) = @_;
+
+	if ($group==1)
+	{
+		$num =~ s/(\d)/ ($1 ? unit($1) :" $zero")."$comma " /eg;
+	}
+	elsif ($group==2)
+	{
+		$num =~ s/(\d)(\d)/ ($1 ? ten($1,$2) : $2 ? " $zero " . unit($2) : " $zero $zero") . "$comma " /eg;
+		$num =~ s/(\d)/ ($1 ? unit($1) :" $zero")."$comma " /e;
+	}
+	elsif ($group==3)
+	{
+		$num =~ s/(\d)(\d)(\d)/ ($1 ? unit($1) :" $zero")." ".($2 ? ten($2,$3) : $3 ? " $zero " . unit($3) : " $zero $zero") . "$comma " /eg;
+		$num =~ s/(\d)(\d)/ ($1 ? ten($1,$2) : $2 ? " $zero " . unit($2) : " $zero $zero") . "$comma " /e;
+		$num =~ s/(\d)/ ($1 ? unit($1) :" $zero")."$comma " /e;
+	}
+	elsif ($num+0==0)
+	{
+		$num = $zero;
+	}
+	else
+	{
+		$num =~ s/\A\s*0+//;
+		my $mill = 0;
+		1 while $num =~ s/(\d)(\d)(\d)(?=\D*\Z)/ hund($1,$2,$3,$mill++,$and) /e;
+		$num =~ s/(\d)(\d)(?=\D*\Z)/ ten($1,$2,$mill)."$comma " /e;
+		$num =~ s/(\d)(?=\D*\Z)/ unit($1,$mill) . "$comma "/e;
+	}
+	return $num;
+}
+
+sub NUMWORDS
+{
+	my $num = shift;
+	my %arg = ( %default_args, @_ );
+	my $group = $arg{group};
+
+	die "Bad chunking option: $group\n" unless $group =~ /\A[0-3]\Z/;
+	my $sign = ($num =~ /\A\s*\+/) ? "plus"
+		 : ($num =~ /\A\s*\-/) ? "minus"
+		 : '';
+
+	my $zero = $arg{zero};
+	my $comma = $arg{comma};
+	my $and = $arg{'and'};
+
+	my @chunks = ($arg{decimal})
+			? $group ? split(/\./, $num) : split(/\./, $num, 2)
+			: ($num);
+
+	my $first = 1;
+
+	if ($chunks[0] eq '') { $first=0; shift @chunks; }
+
+	foreach ( @chunks )
+	{
+		s/\D//g;
+		$_ = '0' unless $_;
+
+		if (!$group && !$first) { $_ = enword($_,1,$zero,$comma,$and) }
+		else         	        { $_ = enword($_,$group,$zero,$comma,$and) }
+
+		s/, \Z//;
+		s/\s+,/,/g;
+		s/, (\S+)\s+\Z/ $and $1/ if !$group and $first;
+		s/\s+/ /g;
+		s/(\A\s|\s\Z)//g;
+		$first = '' if $first;
+	}
+
+	my @numchunks = ();
+	if ($first =~ /0/)
+	{
+		unshift @chunks, '';
+	}
+	else
+	{
+		@numchunks = split /\Q$comma /, $chunks[0];
+	}
+
+	foreach (@chunks[1..$#chunks])
+	{
+		push @numchunks, $arg{decimal};
+		push @numchunks, split /\Q$comma /;
+	}
+
+	if (wantarray)
+	{
+		unshift @numchunks, $sign if $sign;
+		return @numchunks
+	}
+	elsif ($group)
+	{
+		return ($sign?"$sign ":'') .  join ", ", @numchunks;
+	}
+	else
+	{
+		$num = ($sign?"$sign ":'') . shift @numchunks;
+		$first = ($num !~ /$arg{decimal}\Z/);
+		foreach ( @numchunks )
+		{
+			if (/\A$arg{decimal}\Z/)
+			{
+				$num .= " $_";
+				$first = 0;
+			}
+			elsif ($first)
+			{
+				$num .= "$comma $_";
+			}
+			else
+			{
+				$num .= " $_";
+			}
+		}
+		return $num;
+	}
 }
 
 1;
